/*
 * Simple IRCBOT...  
 * By Greg Alexander  <galexand@sietch.bloomington.in.us> 
 */

#include "otools.h"
#include "protos.h"
#include "colors.h"
#include "macros.h"

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/termio.h>
#include <sys/wait.h>
#include <errno.h>
#include <getopt.h>
#include <X11/Xlib.h>
#include <X11/Xlibint.h>

#define X_SIZE width		/* size in the X dimension of window (tree
				 * is full-width) */
#define Y_SIZE (height-20)	/* size in the Y dimension of tree area */
#define X_START_SIZE 860	/* initial size in the X dimension of window */
#define Y_START_SIZE 500	/* initial size in the Y dimension of window */
#define MAX_LINKS 30		/* number of possible links per server */
#define MAX_DEPTH 100		/* Maximum depth of our tree */
#define MAX_TRACES 200		/* number of possible simultaneous traces */
#define MESSAGE_DELAY 300	/* How long to wait between calls to outsock in miliseconds */

#define Gstrcmp(x,y) (gstrcmp(x,y) && gstrcmp(y,x))

struct sockaddr_in remote;
struct hostent *host;
char *server, *sfilename, *selected;
int depth=0;
unsigned int rows[MAX_DEPTH + 1];
char nick[15];
int width, height;
int plug=-1, done=0;
unsigned char do_server=1, save_on_exit=0, check_lag=0, check_links=0, nonblocking_outsock=0;
time_t check_lag_time=0, check_links_time=0;
int global_argc;
char **global_argv;
unsigned long nmt=0;
struct link {
	char a[100], b[100];
	int lag;
	unsigned char old;
	struct link *next;
} *root, *oldroot = NULL;
struct server {
	char *name;
	struct server *uplink, *links[MAX_LINKS];
	int ystart;
	float ideal, distance;
	int x1, y1, x2, y2, nlinks;
	struct link *l;
	unsigned char old;	/* 0 is normal, 1 is old, 2 is new */
} *servers = NULL, *oldservers = NULL;
typedef struct {
	char *dest, *last;
	unsigned long int when;
} tracerec;
tracerec *traces[MAX_TRACES] = {NULL,};
struct sock_queue {
	struct sock_queue *next;
	char *s;
} *sockroot=NULL; 

#define PRIVMSG "PRIVMSG"	/* These #defines are for what the IRC */
#define NOTICE "NOTICE"		/* protocol identifies these actions with... */
#define KICK "KICK"
#define JOIN "JOIN"
#define PART "PART"
#define INVITE "INVITE"
#define TOPIC "TOPIC"
#define SERVERTOPIC "332"
#define WHORESULTS "352"
#define ERRONEUSNICKNAME "432"
#define NICKNAMEINUSE "433"
#define LOGINACCEPTED "001"
#define BANNEDFROMSERV "465"
#define CLOSINGLINK "losin"
#define TRACE "200"
#define LINK "364"
#define DONELINK "365"
#define WHOIS_SERVER "312"

void proc_whois_server(int argc, char *argv[], char *b);
void proc_trace(int argc, char *argv[], char *b);
void proc_eot(int argc, char *argv[], char *b);
void proc_donelink(int argc, char *argv[], char *b);
void proc_link(int argc, char *argv[], char *b);
void proc_newnick(int argc, char *argv[], char *b);
void proc_login(int argc, char *argv[], char *b);
void proc_banned(int argc, char *argv[], char *b);
void proc_close(int argc, char *argv[], char *b);

struct handler {
	char *argv1;
	void (*func)(int argc, char *argv[], char *b);
} handlers[] = {
    { WHOIS_SERVER, proc_whois_server },
    { TRACE, proc_trace },
    { "206", proc_eot },
    { "204", proc_eot },
    { "203", proc_eot },
    { DONELINK, proc_donelink },
    { LINK, proc_link },
    { ERRONEUSNICKNAME, proc_newnick },
    { NICKNAMEINUSE, proc_newnick },
    { LOGINACCEPTED, proc_login },
    { BANNEDFROMSERV, proc_banned },
    { CLOSINGLINK, proc_close },
    { NULL, NULL } };

void print_usage(char *argv0);
unsigned long mtime(void);
void outsock(char *fmt,...);
void add_to_sock_queue(char *s);
void check_sock_queue(void);
int irconnect(char *hostn);
char *randcrap(int numchars);
void process_incoming_line(char *buf);
void select_loop(void);
void irc_select_loop(void);
void add_link(char *a, char *b, int old = 0);
void free_tree(struct server *s);
void build_tree(char *base);
void center_servers(struct server *s, int level = 0);
int build_tree(char *base, int level, struct server *s);
int add_trace(char *target);
void add_tracedata(int i,char *a,char *b);
void update_rows(struct server *s, int level = 0);
void strupr(char *s);
int count_strchr(char *s,char c);
int gstrcmp(char *shit1, char *shit2);
void X_show_tree(void);
void X_show_tree(struct server *s, int level);
void X_show_tree2(void);
struct server *remove_server(struct server *s, char *a, char *b);
void expire_old(struct link **r);
void show_links(struct server *s);
int read_in(char *fn);
int write_out(char *fn);
void rebuild_all(void);
void compare_roots(void);
void add_buttons(void);
int add_button(char *name, void (*func) (Button *, int, int), int x);
void Conf(int x, int y, int w, int h);
void Quit(Button *b, int x, int y);
void View(Button *b, int ix, int y);
void Select(Button *b, int x, int y);
void Lag(Button *b, int x, int y);
void LagAll(struct server *s);
void LagAll(Button *b, int x, int y);
void Links(Button *b, int x, int y);
void ExpireOld(Button * b, int x, int y);
void File1(Button *bb, int x, int y);
void File2(Button *bb, int x, int y);

struct option opts[] = {
	{ "help",0,NULL,'h' },
	{ "server",1,NULL,'s' },
	{ "file",1,NULL,'f' },
	{ "serverfile",1,NULL,'S' },
	{ "checklink",1,NULL,'c' },
	{ "checklag",1,NULL,'C' },
	{ NULL,0,NULL,0 } };

int main(int argc, char **argv)
{
	int i,c;
	char *chost=NULL;
	int option_index;

	printf("ll++ Copyright (c) November, 1996, Greg Alexander.\n");
	printf("To be distributed under the terms of the GNU Copyleft, found in the file\nCOPYING distributed with this program.\n\n");
	printf("Play nice with this toy.  Do not use this for attacking the network,\nthat's why traditional linklookers were outlawed in the first place!\n");

	global_argc = argc;	/* for X initialization calls (adds, i.e., "display" option) */
	global_argv = argv;

	srand(getpid() + getuid() | time(NULL)); /* init rand and nick */
	strcpy(nick, randcrap(9));

	signal(SIGPIPE, SIG_IGN);	/* ignore some usually fatal signals that we don't care about */
	signal(SIGHUP, SIG_IGN);

	while ((c=getopt_long(argc,argv,"hs:f:S:c:C:",opts,&option_index))!=EOF)
		switch(c) {
			case 'h':
				print_usage(argv[0]);
				exit(-1);
			case 'f':
				do_server = 0;
			case 'S':
				save_on_exit=((c!='f')?1:0);
				sfilename=strdup(optarg);
				break;
			case 's':
				chost=strdup(optarg);
				break;
			case 'c':
				check_links = atoi(optarg);
				break;
			case 'C':
				check_lag = atoi(optarg);
				break;
			default:
			case '?':
				fprintf(stderr,"%s: unknown argument!\n",argv[0]);
				print_usage(argv[0]);
				exit(-1);
			case ':':
				fprintf(stderr,"%s: argument needs additional argument!\n",argv[0]);
				print_usage(argv[0]);
				exit(-1);
	 	}

	if (do_server) {
		if (!chost)
			print_usage(argv[0]);
		server=strdup(chost);
		strtok(server, ":");
	}
	if ((!do_server || save_on_exit) && !read_in(sfilename))
		return EXIT_FAILURE;
	if (save_on_exit) {	/* this builds an 'original' tree from file */
		printf("Building tree..."); fflush(stdout);
		build_tree(strdup(server));
		expire_old(&root);
		oldroot=root;
		root=NULL;
		printf("done.\n");
	}

	selected=strdup(server);

	if (do_server) {	/* this connects to our server and grabs our linklist */
		printf("\nConnecting to: %s...", chost); fflush(stdout);
		if (!irconnect(chost))
			exit((fprintf(stderr,"Error connecting to %s!\n",chost),EXIT_FAILURE));
		free(chost);
		printf("done.\nWaiting for login to be accepted..."); fflush(stdout);
		done = 0;
		while (!done)	/* this is a bit cryptic...when the connection
				 * is accepted, it sends WHOIS (me) and LINKS
				 * and stores the data and sets done on end of
				 * LINKS data */
			irc_select_loop();
	}
	if (save_on_exit)
		rebuild_all();
	else {
		printf("Building tree..."); fflush(stdout);
		build_tree(strdup(server));
		printf("done.\n");
	}
	check_lag_time = time(NULL) + 2;	/* we want to check lag starting now */
	check_links_time = time(NULL) + check_links; /* we just did links, so start later */
	X_show_tree();
	select_loop();
}

void print_usage(char *argv0)
{
	fprintf(stderr,"usage: %s [options]\n", argv0);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"-h,--help                   Shows this.\n");
	fprintf(stderr,"-s,--server <server>        Tells which server (with optional :port, 6667\n");
	fprintf(stderr,"            is default) connect to.  Not optional!\n");
	fprintf(stderr,"-f,--file <filename>        Reads a file instead of connecting to a server.\n");
	fprintf(stderr,"-S,--serverfile <filename>  Connects to the server, reads the file, shows\n");
	fprintf(stderr,"            the difference.  Saves the file again on exit (assuming you use\n");
	fprintf(stderr,"            the Quit button to exit).  Handy if you run this all day, exit\n");
	fprintf(stderr,"            at night, then want to see how it's changed when you get back.\n");
	fprintf(stderr,"-c,--checklink <seconds>    Checks links every that many seconds.\n");
	fprintf(stderr,"-C,--checklag <seconds>     Checks lag (like LagAll button) every that many\n");
	fprintf(stderr,"            seconds.  If this value is less than 60, you are dumb.\n");
	exit(EXIT_FAILURE);
}

unsigned long mtime(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

void outsock(char *fmt,...)
{
	va_list ap;
	char s[1001];
	
	va_start(ap,fmt);
	vsnprintf(s,1000,fmt,ap);
	va_end(ap);
	s[1000]=0;
	if (nonblocking_outsock) {
		add_to_sock_queue(s);
		check_sock_queue();
	} else {
		if (mtime()<nmt)
			usleep((nmt-mtime()) * 1000L);
		send(plug, s, strlen(s), 0);
		nmt=mtime()+MESSAGE_DELAY;
	}
}

void add_to_sock_queue(char *s)
{
	struct sock_queue *n=(struct sock_queue *)malloc(sizeof *n), **t;
	if (!n)
		exit((perror("malloc"),-1));
	n->next=NULL;
	n->s=strdup(s);
	if (!n->s)
		exit((perror("strdup"),-1));
	for (t=&sockroot;*t;t=&((*t)->next)) ;
	*t=n;
}

void check_sock_queue(void)
{
	struct sock_queue *t;

	if (!sockroot || mtime() < nmt)
		return;
	send(plug,sockroot->s,strlen(sockroot->s),0);
	nmt=mtime()+MESSAGE_DELAY;
	t=sockroot;
	sockroot=t->next;
	free(t);
}

/* Connects to hostn.  hostn is server.blah.com[:port]  defaults to 6667 
 * 0 = failure, 1 = success
 */
int irconnect(char *hostn)
{
	struct hostent *host_name;
	struct sockaddr_in socketname;
	char *hname,*s;
	int portnum=0;

	hname=strdup(hostn);
	if (!hname)
		return perror("malloc"),0;
	if ((s=strchr(hname, ':'))) {
		portnum=atoi(s+1);
		*s=0;
	}
	if (!portnum)
		portnum = 6667;

	if ((host_name = gethostbyname(hname)) == (struct hostent *)NULL)
		return fprintf(stderr,"Host not found: %s!\n",hname),free(hname),0;
	free(hname);
	if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return perror("socket"),0;
	socketname.sin_family = AF_INET;
	socketname.sin_port = htons(portnum);
	memcpy(&socketname.sin_addr, host_name->h_addr, host_name->h_length);
	signal(SIGALRM, SIG_IGN);
	alarm(60);
	if ((connect(plug, (struct sockaddr *)&socketname, sizeof(socketname))) < 0) {
		if (errno == EINTR)
			fprintf(stderr,"Timed out!\n");
		else
			perror("connect");
		return 0;
	}
	outsock("USER a b c :mIRC user\nNICK %9s\n", nick);
	return 1;
}

/* Returns a string with numchars random characters (a-z) 
 * uses an internal string, so it will overwrite previous results in many cases.
 */
char *randcrap(int numchars)
{
	char randjunk[] = "abcdefghijklmnopqrstuvwxyz";
	static char *s=NULL;
	static int len=0;

	if (len<=numchars) {
		free(s);
		s=(char *)malloc(len=numchars+1);
		if (!s)
			return perror("malloc"),(char *)NULL;
	}
	s[numchars--]=0;
	for (;numchars>=0;numchars--)
		s[numchars] = randjunk[rand()%26];
	return s;
}

void process_incoming_line(char *buf)
{
	char **argv,*b;
	int argc=0,i;

	if (!strncmp(buf, "PING", 4)) {
		outsock("PONG :%s\n", nick);
		return;
	}
	if (!strncmp(buf, "ERROR", 5))
		exit((fprintf(stderr,"\nhrm. %s\n",buf),-1));
	b=strchr(buf+1,':');
	if (b)
		*b++=0;

	argv=(char **)malloc(sizeof(char *)*(count_strchr(buf,' ')+200));
	if (!argv) {
		perror("malloc");
		return;
	}
	argv[argc++]=strtok(buf," ");
	while ((argv[argc++]=strtok(NULL," "))) ;
	argc--;
	if (argc<2)
		return;
	for (i=0;handlers[i].argv1;i++)
		if (!strcmp(handlers[i].argv1,argv[1]))
			handlers[i].func(argc,argv,b);
}

void select_loop(void)
{
	struct timeval t;
	fd_set fds;
	int i;

	o_running=1;
	nonblocking_outsock=1;
	while (o_running) {
		t.tv_sec=0; t.tv_usec=100000;
		FD_ZERO(&fds);
		if (do_server)
			FD_SET(plug,&fds);
		FD_SET(o_dpy->fd,&fds);
		if ((i=select(((o_dpy->fd>plug)?o_dpy->fd:plug)+1,&fds,NULL,NULL,&t))<0)
			perror("select");
		if (do_server && FD_ISSET(plug,&fds))
			irc_select_loop();
		if (FD_ISSET(o_dpy->fd,&fds))
			o_EnterLoop();
		XFlush(o_dpy);
		check_sock_queue();
		if (do_server) {
			if (check_lag && check_lag_time <= time(NULL))
				LagAll((Button *) NULL, 0, 0);
			if (check_links && check_links_time <= time(NULL))
				Links((Button *) NULL, 0, 0);
		}
	}
	o_Quit();
}

void irc_select_loop(void)
{
	int amtread, i, i1;
	char buf[512];
	static int linepos = 0;
	static char linebuf[512];
	struct timeval t;
	fd_set fds;

	t.tv_sec = 0;
	t.tv_usec = 1000;
	FD_ZERO(&fds);
	FD_SET(plug,&fds);
	if ((i1 = select(plug+1, &fds, NULL, NULL, &t))<0) {
		perror("select");
		close(plug);
		do_server=0;
		add_buttons();
		return;
	}
	if (FD_ISSET(plug, &fds)) {
		amtread = recvfrom(plug, buf, 512, 0, (struct sockaddr *)NULL, (int *)0);
		if (amtread < 0) {
			close(plug);
			perror("recvfrom");
			do_server=0;
			add_buttons();
			return;
		}
		for (i = 0; i < amtread; i++) {
			linebuf[linepos] = buf[i];
			if (linebuf[linepos++] == '\n') {
				strtok(linebuf,"\r\n");
				process_incoming_line(linebuf);
				linepos = 0;
			}
		}
	}
}


void add_link(char *a, char *b, int old = 0)
{
	struct link *n, **temp;

	if (!a || !b || !strcasecmp(a, b))
		return;

	n = (struct link *)malloc(sizeof(struct link));
	if (!n)
		exit((perror("malloc"),EXIT_FAILURE));
	strcpy(n->a, a);
	strcpy(n->b, b);
	n->lag=-1;
	n->old=old;
	n->next=(struct link *)NULL;

	for (temp=&root;*temp;temp=&((*temp)->next)) ;
	*temp=n;
}

void free_tree(struct server *s)
{
	int i;

	if (!s)
		return;
	for (i=0; i<s->nlinks; i++)
		free_tree(s->links[i]);
	free(s);
	if (s==servers)
		servers=NULL;
	if (s==oldservers)
		oldservers=NULL;
}

void build_tree(char *base)
{
	int i;

	strtok(base,":");
	if (servers) {
		free_tree(servers);
		expire_old(&root);
	}
	for (i = 1; i < MAX_DEPTH; i++)
		rows[i] = 0;
	rows[0] = 1;
	depth = 1;
	servers = (struct server *)malloc(sizeof(struct server));
	if (!servers)
		exit((perror("malloc"),EXIT_FAILURE));
	servers->name = base;
	servers->uplink = NULL;
	servers->l = NULL;
	servers->ystart = 0;
	servers->old = 0;
	build_tree(base, 1, servers);

	update_rows(servers);

	for (i = 0; i < 3; i++)
		center_servers(servers);
}

#define SWAP(x,y) { \
	temp=s->links[x]; \
	s->links[x]=s->links[y]; \
	s->links[y]=temp; \
	s->links[x]->distance=fabs(s->links[x]->ideal-(float)(x)); \
	s->links[y]->distance=fabs(s->links[y]->ideal-(float)(y)); \
	continue; }

void center_servers(struct server *s, int level = 0)
{
	int i, i1, i2, id, dist;
	struct server *temp;

	if (!s->nlinks) {
		s->distance = 0.0;
		s->ideal = -1.0;
		return;
	}
	for (i = 0; i < s->nlinks; i++)
		center_servers(s->links[i], level + 1);
	if (!s->uplink)
		return;
	s->ideal = ((float)((2 * s->ystart + s->nlinks) * rows[level]) / (float)(2 * rows[level + 1])) - 0.50 - (float)s->uplink->ystart;
	s->distance = fabs(s->ideal - (float)i);
	for (i1 = 0; i1 < s->nlinks; i1++) {
		if (s->links[i1]->distance < .5)
			continue;
		id = (int)(s->links[i1]->ideal + 0.50);
		dist = (int)(s->links[i1]->distance + 0.50);
		if (id < s->nlinks && id >= 0 && s->links[id]->ideal < 0)
			SWAP(i1, id)
		for (i2 = 0; i2 < dist; i2++) {
			if (id - i2 >= 0 && id - i2 < s->nlinks &&
			    s->links[id - i2]->distance < s->links[i1]->distance)
				SWAP(i1, id - i2);
			if (i2 && id + i2 >= 0 && id + i2 < s->nlinks && id + i2 < s->nlinks &&
			    s->links[id + i2]->distance < s->links[i1]->distance)
				SWAP(i1, id + i2);
		}
	}
	update_rows(servers);
}

void mark_old(struct server *s, int old)
{	/* marks the whole subtree old */
	int i;

	s->old = old;
	if (s->l->old==2)
		s->l->old=0;
	for (i = 0; i < s->nlinks; i++)
		mark_old(s->links[i], old);
}

int build_tree(char *base, int level, struct server *s)
{
	struct link *temp;
	char *s2;
	struct server *t;

	if (depth < level)
		depth = level;

	s->nlinks = 0;

	for (temp=root;temp;temp=temp->next) {
		if (Gstrcmp(temp->a,base) && Gstrcmp(temp->b,base))
			continue;
		if (!Gstrcmp(temp->a, base))
			s2 = temp->b;
		else
			s2 = temp->a;
		if (s->uplink && !Gstrcmp(s2, s->uplink->name))
			continue;
		if (level >= MAX_DEPTH)
			return fprintf(stderr, "Maximum depth (%d) exceded from server %s!  Truncated!\n", MAX_DEPTH, base),0;
		rows[level]++;
		if (s->nlinks >= MAX_LINKS)
			return fprintf(stderr, "Gwak!  More than %d links off of %s -- truncating!\n", MAX_LINKS, base),s->nlinks;
		if (temp->old == 1) {
			if (!(t = remove_server(oldservers, s->name, s2)))
				continue;
			s->links[s->nlinks] = t;
			s->links[s->nlinks]->uplink = s;
			mark_old(t, 1);
			s->nlinks++;
		} else {
			s->links[s->nlinks] = (struct server *)malloc(sizeof(struct server));
			if (!s->links[s->nlinks])
				exit((perror("malloc"),EXIT_FAILURE));
			s->links[s->nlinks]->name = s2;
			s->links[s->nlinks]->uplink = s;
			s->links[s->nlinks]->l = temp;
			s->links[s->nlinks]->old = ((temp->old) ? (temp->old) : (s->old));
			build_tree(s2, level + 1, s->links[s->nlinks++]);
		}
	}
	return (s->nlinks);
}

int add_trace(char *target)
{
	tracerec *n;
	int i;

	for (i = 0; i < MAX_TRACES; i++)
		if (traces[i] && !strcasecmp(traces[i]->dest, target))
			return 0;
	n = (tracerec *) malloc(sizeof(tracerec));
	if (!n)
		exit((perror("malloc"),EXIT_FAILURE));

	n->dest=strdup(target);
	if (!n->dest)
		exit((perror("strdup"),EXIT_FAILURE));
	n->last=NULL;
	n->when=0;
	for (i = 0; i<MAX_TRACES && traces[i]; i++) ;
	if (i==MAX_TRACES)
		return fprintf(stderr, "Insufficient space for another trace!\n"), free(n), 0;
	traces[i] = n;
	outsock("TRACE %s\n", target);
	return 1;
}

void add_tracedata(int i,char *a,char *b)
{
	struct link *temp;
	
	if (!traces[i] || !traces[i]->when || !a || !b)
		return;
	for (temp=root;temp;temp=temp->next)
		if ((!Gstrcmp(temp->a, a) && 
		     !Gstrcmp(temp->b, b)) ||
		    (!Gstrcmp(temp->a, b) &&
		     !Gstrcmp(temp->b, a))) {
			temp->lag = time(NULL) - traces[i]->when;
			Conf(1, 1, width,height);
			return;
		}
	fprintf(stderr, "%s -> %s (lag %ld) not found...lag not reported!\n", a, b, time(NULL) - traces[i]->when);
} 

/*
 * Updates rows[] array and the ystart elements of the servers thingy 
 */
void update_rows(struct server *s, int level = 0)
{
	int i;

	if (s == servers && !level) {
		for (i = 1; i < MAX_DEPTH; i++)
			rows[i] = 0;
		rows[0] = 1;
	}
	if (!s)
		return;
	s->ystart = rows[level + 1];
	rows[level + 1] += s->nlinks;
	for (i = 0; i < s->nlinks; i++)
		update_rows(s->links[i], level + 1);
}

void strupr(char *s)
{
	while (*s) {
		*s = toupper(*s);
		s++;
	}
}

int count_strchr(char *s,char c)
{
	int i=0;
	while (*s)
		if (*s++==c)
			i++;
	return i;
}

/*
 * Compares shit1 and shit2 with the '*' wildcards in shit2 recognized, case 
 * insensitive.  returns 0 if they're equal.  otherwise returns 1
 * I'm rather adversely opposed to rewriting this.
 */
int gstrcmp(char *shit1, char *shit2)
{
	char *s[105], *ss, s3[500], s2[500], s1[500];
	int i = 1, i1;

	strcpy(s2, shit2);	/* Take our inputs and put them in internal variables */
	strupr(s2);		/* and strupr them for case insensitivity... */
	strcpy(s1, shit1);
	strupr(s1);
	strcpy(s3, s2);
/*
 * The following stuff takes s2 (from shit2) and parses it into *s[]
 * *s[] contains a bunch of substrings from s2.  For example, if s2 was 
 * "foo*bar" then s[0] = "foo" and s[1] = "bar". 
 */
	s[0] = s2;
	ss = s2;
	while (i < 100) {
		ss = strchr(ss, '*');
		if (!ss)
			break;
		else {
			*ss = 0;
			ss++;
			if (*ss)
				s[i++] = ss;
		}
	}
	if (i == 100) {		/* This happens if we have something with 100 *'s */
		return fprintf(stderr, "What the fuck?  More than 100 *'s in a mask!\n"),1;
	}
	ss = s1;		/* don't ask what the rest of this does */
	if (strncmp(ss, s[0], strlen(s[0])))
		return (1);
	for (i1 = 0; i1 < i; i1++) {
		ss = strstr(ss, s[i1]);
		if (!ss)
			break;
		ss += strlen(s[i1]);
		if (i1 == (i - 1))
			if ((*ss) && (s3[strlen(s3) - 1] != '*'))
				break;
	}
	return (!(i1 == i));
}

void X_show_tree(void)
{
	o_Init("ll++", "ll++", "5x7",	/* App name, titlebar, font */
	       10, 10,			/* Where */
	       X_START_SIZE, Y_START_SIZE, /* size */
	       100, 100,		/* minimum size */
	       3000, 1500,		/* maximum size */
	       C_GRAY,			/* color map */
	       global_argc, global_argv); /* argc/argv */
	o_draw = new Chain();

	width = X_START_SIZE;
	height = Y_START_SIZE;
	X_show_tree(servers, 0);

	o_test = o_draw;

	o_Configure = Conf;

	add_buttons();
}

void X_show_tree(struct server *s, int level)
{
	static Color old("black", "gray40", "gray50", "gray80", "black", "Blue4","black"),
	 notold("black", "gray40", "gray70", "gray80", "black", "Blue4", "black");
	static unsigned int t[MAX_DEPTH];
	static float xscale;
	static char ss[100];
	float yscale;
	int i;

	if (s == servers) {
		for (i = 0; i < depth; i++)
			t[i] = 0;
		xscale = ((float)X_SIZE - 10.0) / (float)depth;
		ss[0] = ss[1] = ' ';
	}
	yscale = ((float)Y_SIZE - 10.0) / (float)rows[level];
	strcpy(ss + 2, s->name);
	ss[(int)((xscale - 20) / 6 + 1)] = 0;
	s->x1 = (int)((float)level * xscale + 5.0);
	s->y1 = (int)((float)t[level] * yscale + 3.0);
	s->x2 = (int)((float)level * xscale + xscale - 15.0);
	s->y2 = (int)((float)t[level] * yscale + yscale);
	o_draw->Add(new Button(s->x1, s->y1,
			       s->x2, s->y2,
			       NONE,
			       (s->old == 0) ? (&o_color) :
			       ((s->old == 1) ? (&old) : (&notold)),
			       0, ss, CLICK, OUT, Select, s->name));
	t[level]++;
	for (i = 0; i < s->nlinks; i++)
		X_show_tree(s->links[i], level + 1);
}

void X_show_tree2(void)
{
	delete o_draw;

	if (!o_running)
		return;

	o_draw = new Chain;
	o_test = o_draw;

	X_show_tree(servers, 0);

	add_buttons();

	o_draw->Draw();
	Conf(1, 1, width, height);
}

struct server *remove_server(struct server *s, char *a, char *b)
{
	int i, i1;
	struct server *t;

	for (i = 0; i < s->nlinks; i++) {
		if (!Gstrcmp(s->name, a) && !Gstrcmp(s->links[i]->name, b)) {
			t = s->links[i];
			for (i1 = i + 1; i1 < s->nlinks; i1++)
				s->links[i1 - 1] = s->links[i1];
			s->nlinks--;
			return t;
		}
		if ((t = remove_server(s->links[i], a, b)))
			return t;
	}
	return NULL;
}

void expire_old(struct link **r)
{
	struct link **temp;

	for (temp = r; *temp;)
		if ((*temp)->old == 1) {
			*temp = (*temp)->next;
		} else
			temp = &((*temp)->next);
}

void show_links(struct server *s)
{
	int i;
	static int yy;
	static char ss[100];
	static Color c("black", "black", "gray60", "gray80", "black", "gray20", "gray100");

	for (i = 0; i < s->nlinks; i++)
		show_links(s->links[i]);
	if (!s->uplink)
		return;
	yy = (s->y1 + s->y2) / 2;
	SetColor(s->l->old ? ((s->l->old == 1) ? c.dark : c.light) : c.hl);
	Line(s->uplink->x2, (s->uplink->y1 + s->uplink->y2) / 2, s->x1, yy);
	if (s->l->lag >= 0) {
		Clear(s->x1 + 3, yy - 3, s->x1 + 13, yy + 3);
		sprintf(ss, "%2d", s->l->lag);
		SetColor(c.fg);
		DrawString(s->x1 + 3, yy + 3, ss);
	}
}

int read_in(char *fn)
{
	FILE *f;
	char s[1000];
	char *a, *b;

	f = fopen(fn, "r");
	if (!f) {
		perror(fn);
		return 0;
	}
	expire_old(&root);

	oldroot = root;
	root = NULL;

	fgets(s, 250, f);
	strtok(s, "\r\n");
	free(server);
	server=strdup(s);
	if (!server)
		exit((perror("strdup"),EXIT_FAILURE));

	while (fgets(s, 1000, f)) {
		strtok(s, "\r\n");
		if (!*s)
			continue;
		a = strtok(s, " \t");
		b = strtok(NULL, " \t");
		add_link(a, b);
		*s = 0;
	}
	fclose(f);
	return 1;
}

int write_out(char *fn)
{
	FILE *f;
	struct link *t;

	f = fopen(fn, "w");
	if (!f) {
		perror(fn);
		return 0;
	}
	fprintf(f, "%s\n", servers->name);
	for (t = root; t; t = t->next) {
		if (t->old == 1)
			continue;
		fprintf(f, "%s %s\n", t->a, t->b);
	}
	fclose(f);
	return 1;
}

void rebuild_all(void)
{
	if (o_running)
		Clear(0, 0, X_SIZE, Y_SIZE + 20);

	printf("Rebuilding tree..."); fflush(stdout);

	expire_old(&oldroot);
	compare_roots();

	oldservers = servers;
	servers = NULL;
	build_tree(strdup(oldservers->name));
	free_tree(oldservers);
	X_show_tree2();
	printf("done.\n");
}

void compare_roots(void)
{
	char *a,*b;
	struct link *temp,*temp2;
	if (!oldroot)
		return;
	for (temp = root; temp; temp = temp->next)
		temp->old = 2;
	for (temp = oldroot; temp; temp = temp->next) {
		a = temp->a;
		b = temp->b;
		for (temp2 = root; temp2; temp2 = temp2->next)
			if ((!strcmp(temp2->a, a) && !strcmp(temp2->b, b)) ||
			    (!strcmp(temp2->a, b) && !strcmp(temp2->b, a))) {
				temp2->old = 0;
				break;
			}
		if (!temp2)
			add_link(a, b, 1);
	}
}

void add_buttons(void)
{
	int x = width - 210;

	Clear(0, height - 20, width - 209, height);

	add_button("Quit", Quit, 60);
	x-=add_button("View", View, x);
	if (do_server) {
		x-=add_button("Lag", Lag, x);
		x-=add_button("LagAll", LagAll, x);
		x-=add_button("Links", Links, x);
	}
	x-=add_button("ExpireOld", ExpireOld, x);
	x-=add_button("->File", File2, x);
	x-=add_button("File->", File1, x);
}

int add_button(char *name, void (*func) (Button *, int, int), int x)
{
	int l = strlen(name) * 7 + 10;

	o_draw->Add(new Button(x - l, height - 20, x, height - 5, NONE, &o_color, 0, name, CLICK, OUT, func, name));
	return l;
}

void Conf(int x, int y, int w, int h)
{
	if (w != width || h != height) {
		width = w;
		height = h;
		Clear(0, 0, width, height);
		X_show_tree2();
	}
	Clear(w - 200, h - 20, w, h - 5);
	SetColor(0);
	DrawString(w - 200, h - 10, selected);

	show_links(servers);
}

void Quit(Button * b, int x, int y)
{
	if (save_on_exit) {
		printf("Saving %s...", sfilename); fflush(stdout);
		if (write_out(sfilename))
			printf("done.\n");
	}
	o_running = 0;
}

void View(Button * b, int x, int y)
{
	Clear(0, 0, X_SIZE, Y_SIZE + 20);

	printf("Rebuilding tree...");
	fflush(stdout);
	build_tree(strdup(selected));
	printf("done.\n");
	X_show_tree2();
}

void Select(Button * b, int x, int y)
{
	free(selected);
	selected=strdup(b->fullmessage);
	if (!selected)
		exit((perror("strdup"),EXIT_FAILURE));
	Conf(1, 1, width,height);
}

void Lag(Button * b, int x, int y)
{
	add_trace(selected);
}

void LagAll(struct server *s)
{
	if (!s->nlinks)
		add_trace(s->name);
	else {
		int i;

		for (i = 0; i < s->nlinks; i++)
			LagAll(s->links[i]);
	}
}

void LagAll(Button *b, int xx, int yy)
{
	LagAll(servers);
	if (check_lag)
		check_lag_time = time(NULL) + check_lag;
}

void Links(Button *b, int x, int y)
{
	printf("Re-grabbing links...");
	fflush(stdout);
	expire_old(&root);
	oldroot = root;
	root = NULL;
	outsock("LINKS\n");
	if (check_links)
		check_links_time = time(NULL) + check_links;
}

void ExpireOld(Button *b, int x, int y)
{
	printf("Rebuilding tree...");
	fflush(stdout);
	expire_old(&root);
	build_tree(strdup(server));
	Clear(0, 0, width, height - 20);
	X_show_tree2();
	printf("done.\n");
}

void File1(Button *bb, int x, int y)
{
	char s[1000];

	printf("File to read? ");
	fflush(stdout);
	fgets(s, 1000, stdin);
	strtok(s, "\r\n");

	if (read_in(s))
		rebuild_all();
}

void File2(Button *bb, int x, int y)
{
	char s[1000];

	printf("File to write? ");
	fflush(stdout);
	fgets(s, 1000, stdin);
	strtok(s, "\r\n");

	printf("Writing...");
	fflush(stdout);
	if (write_out(s))
		printf("done.\n");
}

void proc_whois_server(int argc, char *argv[], char *b)
{
	if (argc<5) {
		fprintf(stderr,"Not enough args from server in proc_whois_server!\n");
		return;
	}
	free(server);
	server=strdup(argv[4]);
	if (!server)
		exit((perror("strdup"),EXIT_FAILURE));
}

void proc_trace(int argc, char *argv[], char *b)
{
	struct link *temp;
	int i;

	if (argc<6) {
		fprintf(stderr,"Not enough args from server in proc_trace!\n");
		return;
	}
	for (i = 0; i < MAX_TRACES; i++)
		if (traces[i] && !strcmp(traces[i]->dest, argv[5]))
			break;
	if (i == MAX_TRACES) {
		fprintf(stderr, "Rcv'd link from \"%s\" with dest \"%s\" that was not req'd!\n", argv[0], argv[5]);
		return;
	}
	argv[0]++;		/* Skip the : at the beginning */
	add_tracedata(i,traces[i]->last,argv[0]);
	free(traces[i]->last);
	traces[i]->last=strdup(argv[0]);
	traces[i]->when=time(NULL);
}

void proc_eot(int argc, char *argv[], char *b)
{
	int i;

	argv[0]++;
	for (i=0;i<MAX_TRACES;i++)
		if (traces[i] && !Gstrcmp(traces[i]->dest, argv[0]))
			break;
	if (i==MAX_TRACES)
		return;
	add_tracedata(i,traces[i]->last,argv[0]);
	free(traces[i]->dest);
	free(traces[i]->last);
	free(traces[i]);
	traces[i]=NULL;
}

void proc_donelink(int argc, char *argv[], char *b)
{
	printf("done.\n");
	if (done)
		rebuild_all();
	else
		done=1;
}

void proc_link(int argc, char *argv[], char *b)
{
	if (argc<5) {
		fprintf(stderr,"Not enough args from server in proc_link!\n");
		return;
	}
	printf(".");
	fflush(stdout);
	add_link(argv[3], argv[4]);
}

void proc_newnick(int argc, char *argv[], char *b)
{
	strcpy(nick,randcrap(9));
	outsock("NICK %s\n",nick);
}

void proc_login(int argc, char *argv[], char *b)
{
	printf("done.\nGrabbing links...");
	fflush(stdout);
	outsock("MODE %s +i\n",nick);
	outsock("WHOIS %s\n",nick);
	outsock("LINKS\n");
}

void proc_banned(int argc, char *argv[], char *b)
{
	exit((fprintf(stderr, "Banned from server!\n"),EXIT_FAILURE));
}

void proc_close(int argc, char *argv[], char *b)
{
	fprintf(stderr,"Other end closed link?!\n");
	close(plug);
	do_server=0;
	add_buttons();
}
